import os
import sys
import hashlib

from cryptography.hazmat.primitives import ciphers, serialization
from cryptography.hazmat.primitives.ciphers import algorithms, modes
from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.backends import default_backend

backend = default_backend()

PY3 = sys.version_info > (3,)

class Cipher:
    def __init__(self):
        self.disable()

    def enable(self, key):
        cipher = ciphers.Cipher(algorithms.AES(key), modes.CFB8(key), backend=backend)
        self.encryptor = cipher.encryptor()
        self.decryptor = cipher.decryptor()

    def disable(self):
        self.encryptor = None
        self.decryptor = None

    def encrypt(self, data):
        if self.encryptor:
            return self.encryptor.update(data)
        else:
            return data

    def decrypt(self, data):
        if self.decryptor:
            return self.decryptor.update(data)
        else:
            return data


def make_keypair():
    return rsa.generate_private_key(
        public_exponent=65537,
        key_size=1024,
        backend=default_backend())

def make_server_id():
    data = os.urandom(10)
    if PY3:
        parts = ["%02x" % c for c in data]
    else:
        parts = ["%02x" % ord(c) for c in data]

    return "".join(parts)

def make_verify_token():
    return os.urandom(4)

def make_shared_secret():
    return os.urandom(16)

def make_digest(*data):
    sha1 = hashlib.sha1()
    for d in data: sha1.update(d)

    digest = int(sha1.hexdigest(), 16)
    if digest >> 39*4 & 0x8:
        return"-%x" % ((-digest) & (2**(40*4)-1))
    else:
        return "%x" % digest

def export_public_key(keypair):
    return keypair.public_key().public_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PublicFormat.SubjectPublicKeyInfo)

def import_public_key(data):
    return serialization.load_der_public_key(
        data=data,
        backend=default_backend())

def encrypt_secret(public_key, shared_secret):
    return public_key.encrypt(
        plaintext=shared_secret,
        padding=padding.PKCS1v15())

def decrypt_secret(keypair, data):
    return keypair.decrypt(
        ciphertext=data,
        padding=padding.PKCS1v15())


import collections
import functools
import gzip

from quarry.types.buffer import Buffer

_kinds = {}
_ids = {}


# Base types --------------------------------------------------------------------------------------

@functools.total_ordering
class _Tag(object):
    def __init__(self, value):
        self.value = value

    @classmethod
    def from_buff(cls, buff):
        raise NotImplementedError

    def to_bytes(self):
        raise NotImplementedError

    def to_obj(self):
        return self.value

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self.value)

    def __eq__(self, other):
        return self.to_obj() == other.to_obj()

    def __lt__(self, other):
        return self.to_obj() < other.to_obj()


class _DataTag(_Tag):
    fmt = None

    @classmethod
    def from_buff(cls, buff):
        return cls(buff.unpack(cls.fmt))

    def to_bytes(self):
        return Buffer.pack(self.fmt, self.value)


class _ArrayTag(_Tag):
    fmt = None

    @classmethod
    def from_buff(cls, buff):
        array_length = buff.unpack('i')
        return cls(list(buff.unpack(cls.fmt*array_length)))

    def to_bytes(self):
        return (
            Buffer.pack('i', len(self.value)) +
            Buffer.pack(self.fmt*len(self.value), *self.value))


# NBT tags ----------------------------------------------------------------------------------------

class TagByte(_DataTag):
    fmt = 'b'


class TagShort(_DataTag):
    fmt = 'h'


class TagInt(_DataTag):
    fmt = 'i'


class TagLong(_DataTag):
    fmt = 'q'


class TagFloat(_DataTag):
    fmt = 'f'


class TagDouble(_DataTag):
    fmt = 'd'


class TagString(_Tag):

    @classmethod
    def from_buff(cls, buff):
        string_length = buff.unpack('H')
        return cls(buff.read(string_length).decode('utf8'))

    def to_bytes(self):
        data = self.value.encode('utf8')
        return Buffer.pack('H', len(data)) + data


class TagByteArray(_ArrayTag):
    fmt = 'b'


class TagIntArray(_ArrayTag):
    fmt = 'i'


class TagList(_Tag):
    @classmethod
    def from_buff(cls, buff):
        inner_kind_id = buff.unpack('b')
        inner_kind = _kinds[inner_kind_id]
        array_length = buff.unpack('i')
        if isinstance(inner_kind, _DataTag):
            return cls(list(buff.unpack(inner_kind.fmt * array_length)))
        else:
            return cls([inner_kind.from_buff(buff) for _ in range(array_length)])

    def to_bytes(self):
        if len(self.value) > 0:
            type_id = _ids[type(self.value[0])]
        else:
            type_id = 1
        return (
            Buffer.pack('b', type_id) +
            Buffer.pack('i', len(self.value)) +
            b"".join(tag.to_bytes() for tag in self.value))

    def to_obj(self):
        return [tag.to_obj() for tag in self.value]


class TagCompound(_Tag):
    root = False
    preserve_order = False

    @classmethod
    def from_buff(cls, buff):
        if cls.preserve_order:
            value = collections.OrderedDict()
        else:
            value = {}

        while True:
            kind_id = buff.unpack('b')
            if kind_id == 0:
                return cls(value)
            kind = _kinds[kind_id]
            name = TagString.from_buff(buff).value
            tag = kind.from_buff(buff)
            value[name] = tag
            if cls.root:
                return cls(value)

    def to_bytes(self):
        string = b""
        for name, tag in self.value.items():
            string += Buffer.pack('b', _ids[type(tag)])
            string += TagString(name).to_bytes()
            string += tag.to_bytes()

        if len(self.value) == 0 or not self.root:
            string += Buffer.pack('b', 0)

        return string

    def to_obj(self):
        return dict((name, tag.to_obj()) for name, tag in self.value.items())


class TagRoot(TagCompound):
    root = True


# Register tags -----------------------------------------------------------------------------------

_kinds[0] = type(None)
_kinds[1] = TagByte
_kinds[2] = TagShort
_kinds[3] = TagInt
_kinds[4] = TagLong
_kinds[5] = TagFloat
_kinds[6] = TagDouble
_kinds[7] = TagByteArray
_kinds[8] = TagString
_kinds[9] = TagList
_kinds[10] = TagCompound
_kinds[11] = TagIntArray
_ids.update({v: k for k, v in _kinds.items()})

# Files -------------------------------------------------------------------------------------------

class NBTFile(object):
    root_tag = None

    def __init__(self, root_tag):
        self.root_tag = root_tag

    @classmethod
    def load(cls, path):
        with gzip.open(path, 'rb') as fd:
            buff = Buffer()
            buff.add(fd.read())
            return cls(TagRoot.from_buff(buff))

    def save(self, path):
        with gzip.open(path, 'wb') as fd:
            fd.write(self.root_tag.to_bytes())

# Debug --------------------------------------------------------------------------------------------

def alt_repr(tag, level=0):
    """
    Returns a human-readable representation of a tag using the same format as
    used the NBT specification.
    """
    name = lambda kind: type(kind).__name__.replace("Tag", "TAG_")

    if isinstance(tag, _ArrayTag):
        return "%s%s: %d entries" % (
            "  " * level,
            name(tag),
            len(tag.value))

    elif isinstance(tag, TagList):
        return "%s%s: %d entries\n%s{\n%s\n%s}" % (
            "  " * level,
            name(tag),
            len(tag.value),
            "  " * level,
            u"\n".join(alt_repr(tag, level+1) for tag in tag.value),
            "  " * level)

    elif isinstance(tag, TagRoot):
        return u"\n".join(
                alt_repr(tag, level).replace(': ', '("%s"): ' % name, 1)
                for name, tag in tag.value.items())

    elif isinstance(tag, TagCompound):
        return "%s%s: %d entries\n%s{\n%s\n%s}" % (
            "  " * level,
            name(tag),
            len(tag.value),
            "  " * level,
            u"\n".join(
                alt_repr(tag, level+1).replace(': ', '("%s"): ' % name, 1)
                for name, tag in tag.value.items()),
            "  " * level)

    elif isinstance(tag, TagString):
        return '%s%s: "%s"' % (
            "  " * level,
            name(tag),
            tag.value)

    else:
        return "%s%s: %r" % (
            "  " * level,
            name(tag),
            tag.value)



import functools
import re

from quarry.types.buffer import Buffer

try:
  basestring
except NameError:
  basestring = str

def _load_styles():
    data = {
        "0": "black",
        "1": "dark_blue",
        "2": "dark_green",
        "3": "dark_aqua",
        "4": "dark_red",
        "5": "dark_purple",
        "6": "gold",
        "7": "gray",
        "8": "dark_gray",
        "9": "blue",
        "a": "green",
        "b": "aqua",
        "c": "red",
        "d": "light_purple",
        "e": "yellow",
        "f": "white",
        "k": "obfuscated",
        "l": "bold",
        "m": "strikethrough",
        "n": "underline",
        "o": "italic",
        "r": "reset",
    }

    code_by_name = {}
    code_by_prop = {}
    for code, name in data.items():
        code_by_name[name] = code
        if code in "klmnor":
            if name == "underline":
                prop = "underlined"
            else:
                prop = name
            code_by_prop[prop] = code

    return code_by_name, code_by_prop

code_by_name, code_by_prop = _load_styles()

@functools.total_ordering
class Message(object):
    """
    Represents a Minecraft chat message.
    """
    def __init__(self, value):
        self.value = value

    @classmethod
    def from_buff(cls, buff):
        return cls(buff.unpack_json())

    def to_bytes(self):
        return Buffer.pack_json(self.value)

    @classmethod
    def from_string(cls, string):
        return cls({'text': string})

    def to_string(self, strip_styles=True):
        """
        Minecraft uses a JSON format to represent chat messages; this method
        retrieves a plaintext representation, optionally including styles
        encoded using old-school chat codes (U+00A7 plus one character).
        """

        def parse(obj):
            if isinstance(obj, basestring):
                return obj
            if isinstance(obj, list):
                return "".join((parse(e) for e in obj))
            if isinstance(obj, dict):
                text = ""
                for prop, code in code_by_prop.items():
                    if obj.get(prop):
                        text += u"\u00a7" + code
                if "color" in obj:
                    text += u"\u00a7" + code_by_name[obj["color"]]
                if "translate" in obj:
                    text += obj["translate"]
                    if "with" in obj:
                        args = ", ".join((parse(e) for e in obj["with"]))
                        text += "{%s}" % args
                if "text" in obj:
                    text += obj["text"]
                if "extra" in obj:
                    text += parse(obj["extra"])
                return text

        text = parse(self.value)
        if strip_styles:
            text = self.strip_chat_styles(text)
        return text

    @classmethod
    def strip_chat_styles(cls, text):
        return re.sub(u"\u00A7.", "", text)

    def __eq__(self, other):
        return self.value == other.value

    def __lt__(self, other):
        return self.value < other.value

    def __str__(self):
        return self.to_string()


from twisted.internet.task import LoopingCall


class Task(object):
    ticker = None

    def stop(self):
        self.ticker.remove(self)


class LoopTask(Task):
    def __init__(self, ticker, interval, callback):
        self.ticker = ticker
        self.interval = interval
        self.callback = callback

    def update(self):
        if self.ticker.tick % self.interval == 0:
            self.callback()


class DelayTask(Task):
    def __init__(self, ticker, delay, callback):
        self.ticker = ticker
        self.delay = delay
        self.callback = callback
        self.restart()

    def restart(self):
        self.target = self.ticker.tick + self.delay

    def update(self):
        if self.ticker.tick >= self.target:
            self.callback()
            self.stop()


class Ticker(object):
    #: The current tick
    tick = 0

    #: Interval between ticks, in seconds
    interval = 1.0/20

    #: Maximum number of delayed ticks before they're all skipped
    max_lag = 40

    stopped = False

    def __init__(self, logger):
        self._logger = logger
        self._tasks = []
        self._impl = LoopingCall.withCount(self._update)

    def start(self):
        """
        Start running the tick loop.
        """
        self._impl.start(self.interval, now=False)

    def stop(self):
        """
        Stop running the tick loop.
        """
        if self.stopped:
            return

        self._impl.stop()
        self.stopped = True

    def add_loop(self, interval, callback):
        """
        Repeatedly run a callback.

        :param interval: The interval in ticks
        :param callback: The callback to run
        :return: An instance providing a ``stop()`` method
        """
        task = LoopTask(self, interval, self._wrap(callback))
        self._tasks.append(task)
        return task

    def add_delay(self, delay, callback):
        """
        Run a callback after a delay.

        :param delay: The delay in ticks
        :param callback: The callback to run
        :return: An instance providing ``stop()`` and ``restart()`` methods
        """
        task = DelayTask(self, delay, self._wrap(callback))
        self._tasks.append(task)
        return task

    def remove(self, task):
        """
        Removes a task, effectively cancelling it.

        :param task: The task to remove
        """
        self._tasks.remove(task)

    def remove_all(self):
        """
        Removes all registered tasks, effectively cancelling them.
        """
        del self._tasks[:]

    def _update(self, count):
        if count >= self.max_lag:
            self._logger.warn("Can't keep up! Skipping %d ticks" % (count - 1))
            count = 1

        for _ in range(count):
            for task in list(self._tasks):
                task.update()
            self.tick += 1

    def _wrap(self, callback):
        def fn():
            try:
                callback()
            except Exception as e:
                self._logger.exception(e)
        return fn


import string
import sys
import logging
import zlib
from twisted.internet import protocol

from quarry.data import packets
from quarry.types.buffer import Buffer, BufferUnderrun
from quarry.net.crypto import Cipher
from quarry.net.ticker import Ticker

PY3 = sys.version_info > (3,)

logging.basicConfig(format="%(name)s | %(levelname)s | %(message)s")

protocol_modes = {
    0: 'init',
    1: 'status',
    2: 'login',
    3: 'play'
}
protocol_modes_inv = dict(((v, k) for k, v in protocol_modes.items()))


class ProtocolError(Exception):
    pass


class PacketDispatcher(object):
    def dispatch(self, lookup_args, buff):
        handler = getattr(self, "packet_%s" % "_".join(lookup_args), None)
        if handler is not None:
            handler(buff)
            return True
        return False

    def dump_packet(self, data):
        lines = ['Packet dump:']
        bytes_read = 0
        while len(data) > 0:
            data_line, data = data[:16], data[16:]

            l_hex = []
            l_str = []
            for i, c in enumerate(data_line):
                if PY3:
                    l_hex.append("%02x" % c)
                    c_str = data_line[i:i+1]
                    l_str.append(c_str if c_str in string.printable else ".")
                else:
                    l_hex.append("%02x" % ord(c))
                    l_str.append(c if c in string.printable else ".")

            l_hex.extend(['  '] * (16 - len(l_hex)))
            l_hex.insert(8, '')

            lines.append("%08x  %s  |%s|" % (
                bytes_read,
                " ".join(l_hex),
                "".join(l_str)))

            bytes_read += len(data_line)

        return "\n    ".join(lines + ["%08x" % bytes_read])


class Protocol(protocol.Protocol, PacketDispatcher, object):
    """Shared logic between the client and server"""

    #: Usually a reference to the :class:`Buffer` class. This is useful when
    #: constructing a packet payload for use in :meth:`send_packet`
    buff_type = None

    #: The logger for this protocol.
    logger = None

    #: A reference to a :class:`Ticker` instance.
    ticker = None

    #: A reference to the factory
    factory = None

    #: The IP address of the remote.
    remote_addr = None

    recv_direction = None
    send_direction = None
    protocol_version = packets.default_protocol_version
    protocol_mode = "init"
    compression_threshold = None
    compression_enabled = False
    in_game = False
    closed = False

    def __init__(self, factory, remote_addr):
        self.factory = factory
        self.remote_addr = remote_addr

        self.buff_type = self.factory.buff_type
        self.recv_buff = self.buff_type()
        self.cipher = Cipher()

        self.logger = logging.getLogger("%s{%s}" % (
            self.__class__.__name__,
            self.remote_addr.host))
        self.logger.setLevel(self.factory.log_level)

        self.ticker = self.factory.ticker_type(self.logger)
        self.ticker.start()

        self.connection_timer = self.ticker.add_delay(
            delay=self.factory.connection_timeout / self.ticker.interval,
            callback=self.connection_timed_out)

        self.setup()

    ### Fix ugly twisted methods ----------------------------------------------

    def dataReceived(self, data):
        return self.data_received(data)

    def connectionMade(self):
        return self.connection_made()

    def connectionLost(self, reason=None):
        return self.connection_lost(reason)

    ### Convenience functions -------------------------------------------------

    def check_protocol_mode_switch(self, mode):
        transitions = [
            ("init", "status"),
            ("init", "login"),
            ("login", "play")
        ]

        if (self.protocol_mode, mode) not in transitions:
            raise ProtocolError("Cannot switch protocol mode from %s to %s"
                                % (self.protocol_mode, mode))

    def switch_protocol_mode(self, mode):
        self.check_protocol_mode_switch(mode)
        self.protocol_mode = mode

    def set_compression(self, compression_threshold):
        if not self.compression_enabled:
            self.compression_enabled = True
            self.logger.debug("Compression enabled")

        self.compression_threshold = compression_threshold
        self.logger.debug("Compression threshold set to %d bytes" % compression_threshold)

    def close(self, reason=None):
        """Closes the connection"""

        if not self.closed:
            if reason:
                reason = "Closing connection: %s" % reason
            else:
                reason = "Closing connection"

            if self.in_game:
                self.logger.info(reason)
            else:
                self.logger.debug(reason)

            self.transport.loseConnection()
            self.closed = True

    def log_packet(self, prefix, name):
        """Logs a packet at debug level"""

        self.logger.debug("Packet %s %s/%s" % (
            prefix,
            self.protocol_mode,
            name))

    ### General callbacks -----------------------------------------------------

    def setup(self):
        """Called when the Protocol's initialiser is finished"""

        pass

    def protocol_error(self, err):
        """Called when a protocol error occurs"""

        self.logger.exception(err)
        self.close("Protocol error")

    ### Connection callbacks --------------------------------------------------

    def connection_made(self):
        """Called when the connection is established"""

        self.logger.debug("Connection made")

    def connection_lost(self, reason=None):
        """Called when the connection is lost"""

        self.closed = True
        if self.in_game:
            self.player_left()
        self.logger.debug("Connection lost")

        self.ticker.stop()

    def connection_timed_out(self):
        """Called when the connection has been idle too long"""

        self.close("Connection timed out")

    ### Auth callbacks --------------------------------------------------------

    def auth_ok(self, data):
        """Called when auth with mojang succeeded (online mode only)"""

        pass

    def auth_failed(self, err):
        """Called when auth with mojang failed (online mode only)"""

        self.logger.warn("Auth failed: %s" % err.value)
        self.close("Auth failed: %s" % err.value)

    ### Player callbacks ------------------------------------------------------

    def player_joined(self):
        """Called when the player joins the game"""

        self.in_game = True

    def player_left(self):
        """Called when the player leaves the game"""

        pass

    ### Packet handling -------------------------------------------------------

    def data_received(self, data):
        # Decrypt data
        data = self.cipher.decrypt(data)

        # Add it to our buffer
        self.recv_buff.add(data)

        # Read some packets
        while not self.closed:
            # Save the buffer, in case we read an incomplete packet
            self.recv_buff.save()

            # Try to read a packet
            try:
                max_bits = 32 if self.protocol_mode == "play" else 21
                packet_length = self.recv_buff.unpack_varint(max_bits=max_bits)
                packet_body = self.recv_buff.read(packet_length)

            # Incomplete packet read, restore the buffer.
            except BufferUnderrun:
                self.recv_buff.restore()
                break

            # Load the packet body into a buffer
            packet_buff = self.buff_type()
            packet_buff.add(packet_body)

            try:  # Catch protocol errors
                try:  # Catch buffer overrun/underrun
                    if self.compression_enabled:
                        uncompressed_length = packet_buff.unpack_varint()

                        if uncompressed_length > 0:
                            data = zlib.decompress(packet_buff.read())
                            packet_buff = Buffer()
                            packet_buff.add(data)
                    ident = packet_buff.unpack_varint()
                    key = (
                        self.protocol_version,
                        self.protocol_mode,
                        self.recv_direction,
                        ident)
                    try:
                        name = packets.packet_names[key]
                    except KeyError:
                        raise ProtocolError("No name known for packet: %s"
                                            % (key,))
                    self.packet_received(packet_buff, name)

                except BufferUnderrun:
                    raise ProtocolError("Packet is too short!")

                if len(packet_buff) > 0:
                    raise ProtocolError("Packet is too long!")

            except ProtocolError as e:
                self.protocol_error(e)
                break

            # We've read a complete packet, so reset the inactivity timeout
            self.connection_timer.restart()

    def packet_received(self, buff, name):
        """
        Called when a packet is received from the remote. Usually this method
        dispatches the packet to a method named ``packet_<packet name>``, or
        calls :meth:`packet_unhandled` if no such methods exists. You might
        want to override this to implement your own dispatch logic or logging.
        """

        self.log_packet(". recv", name)

        dispatched = self.dispatch((name,), buff)

        if not dispatched:
            self.packet_unhandled(buff, name)

    def packet_unhandled(self, buff, name):
        """
        Called when a packet is received that is not hooked. The default
        implementation silently discards the packet.
        """

        buff.discard()

    def send_packet(self, name, *data):
        """Sends a packet to the remote."""

        if self.closed:
            return

        self.log_packet("# send", name)

        data = b"".join(data)

        # Prepend ident
        key = (
            self.protocol_version,
            self.protocol_mode,
            self.send_direction,
            name)
        try:
            ident = packets.packet_idents[key]
        except KeyError:
            raise ProtocolError("No ID known for packet: %s" % (key,))
        data = Buffer.pack_varint(ident) + data

        if self.compression_enabled:
            # Compress data and prepend uncompressed data length
            if len(data) >= self.compression_threshold:
                data = Buffer.pack_varint(len(data)) + zlib.compress(data)
            else:
                data = Buffer.pack_varint(0) + data

        # Prepend packet length
        max_bits = 32 if self.protocol_mode == "play" else 21
        data = self.buff_type.pack_varint(len(data), max_bits=max_bits) + data

        # Encrypt
        data = self.cipher.encrypt(data)

        # Send
        self.transport.write(data)


class Factory(protocol.Factory, object):
    protocol = Protocol
    buff_type = Buffer
    ticker_type = Ticker
    log_level = logging.INFO
    connection_timeout = 30

    minecraft_versions = packets.minecraft_versions

    def buildProtocol(self, addr):
        return self.protocol(self, addr)


import struct
import json

from quarry.types.uuid import UUID
from quarry.types.chunk import BlockArray, LightArray

# Python 3 compat
try:
    basestring
except NameError:
    basestring = str
try:
    xrange
except NameError:
    xrange = range


class BufferUnderrun(Exception):
    pass


class Buffer(object):
    buff = b""
    pos = 0

    def __init__(self, data=None):
        if data:
            self.buff = data

    def __len__(self):
        return len(self.buff) - self.pos

    def add(self, data):
        """
        Add some bytes to the end of the buffer.
        """

        self.buff += data

    def save(self):
        """
        Saves the buffer contents.
        """

        self.buff = self.buff[self.pos:]
        self.pos = 0

    def restore(self):
        """
        Restores the buffer contents to its state when :meth:`save` was last
        called.
        """

        self.pos = 0

    def discard(self):
        """
        Discards the entire buffer contents.
        """

        self.pos = len(self.buff)

    def read(self, length=None):
        """
        Read *length* bytes from the beginning of the buffer, or all bytes if
        *length* is ``None``
        """

        if length is None:
            data = self.buff[self.pos:]
            self.pos = len(self.buff)
        else:
            if self.pos + length > len(self.buff):
                raise BufferUnderrun()

            data = self.buff[self.pos:self.pos+length]
            self.pos += length

        return data

    def unpack(self, fmt):
        """
        Unpack a struct. The format accepted is the same as for
        ``struct.unpack()``.
        """

        fmt = ">"+fmt
        length = struct.calcsize(fmt)
        fields = struct.unpack(fmt, self.read(length))
        if len(fields) == 1:
            fields = fields[0]
        return fields

    def unpack_string(self):
        """
        Unpack a Minecraft string (varint-prefixed utf8).
        """

        length = self.unpack_varint(max_bits=16)
        text = self.read(length).decode("utf-8")
        return text

    def unpack_json(self):
        """
        Unpack a Minecraft string and interpret it as JSON.
        """

        obj = json.loads(self.unpack_string())
        return obj

    def unpack_chat(self):
        """
        Unpack a Minecraft chat message.
        """
        from quarry.types import chat
        return chat.Message.from_buff(self)

    def unpack_varint(self, max_bits=32, signed=False):
        """
        Unpacks a varint.
        """

        number = 0
        for i in range(5):
            b = self.unpack("B")
            number |= (b & 0x7F) << 7*i
            if not b & 0x80:
                break

        if number & (1<<31):
            number -= 1<<32
            if not signed:
                raise ValueError("varint cannot be negative: %d" % number)

        number_min = -1 << (max_bits - 1)
        number_max = +1 << (max_bits - 1)
        if not (number_min <= number < number_max):
            raise ValueError("varint does not fit in range: %d <= %d < %d"
                             % (number_min, number, number_max))

        return number

    def unpack_uuid(self):
        """
        Unpacks a UUID.
        """

        return UUID.from_bytes(self.read(16))

    def unpack_position(self):
        """
        Unpacks a position.
        """

        def unpack_twos_comp(bits, number):
            if (number & (1 << (bits - 1))) != 0:
                number = number - (1 << bits)
            return number

        number = self.unpack('Q')
        x = unpack_twos_comp(26, (number >> 38))
        y = unpack_twos_comp(12, (number >> 26 & 0xFFF))
        z = unpack_twos_comp(26, (number & 0x3FFFFFF))
        return x, y, z

    def unpack_slot(self):
        """
        Unpacks a slot.
        """

        slot = {}
        slot['id'] = self.unpack('h')
        if slot['id'] != -1:
            slot['count'] = self.unpack('b')
            slot['damage'] = self.unpack('h')
            slot['tag'] = self.unpack_nbt()

        return slot

    def unpack_nbt(self):
        """
        Unpacks NBT tag(s).
        """

        from quarry.types import nbt
        return nbt.TagRoot.from_buff(self)

    def unpack_chunk_section(self, overworld=True):
        """
        Unpacks a chunk section. Returns a 3-tuple of
        ``(blocks, block_lights, sky_lights)``, where *sky_lights* is ``None``
        when *overworld* is ``False``. The returned values are sequences of
        length 4096 (16x16x16).
        """
        bits = self.unpack('B')
        if bits < 4:   bits = 4
        elif bits > 8: bits = 13

        palette = [self.unpack_varint() for _ in xrange(self.unpack_varint())]
        blocks = BlockArray(self.unpack('Q' * self.unpack_varint()), bits, palette)
        block_lights = LightArray(self.unpack('B' * 2048))
        if overworld:
            sky_lights = LightArray(self.unpack('B' * 2048))
        else:
            sky_lights = None

        return blocks, block_lights, sky_lights


    def unpack_entity_metadata_1_7(self):
        """
        Unpacks entity metadata. Compatible with Minecraft 1.7 and 1.8.
        """
        metadata = []
        while True:
            b = self.unpack('B')
            if b == 127:
                return metadata
            ty, key = b >> 5, b & 0x1F
            if   ty == 0: val = self.unpack('b')
            elif ty == 1: val = self.unpack('h')
            elif ty == 2: val = self.unpack('i')
            elif ty == 3: val = self.unpack('f')
            elif ty == 4: val = self.unpack_string()
            elif ty == 5: val = self.unpack_slot()
            elif ty == 6: val = self.unpack('iii')
            elif ty == 7: val = self.unpack('fff')
            else: raise ValueError("Unknown entity metadata type: %d" % ty)
            metadata.append((ty, key, val))

    def unpack_entity_metadata(self):
        """
        Unpacks entity metadata. Compatible with Minecraft 1.9+.
        """
        def unpack_optional(unpacker):
            if self.unpack('?'):
                return unpacker()
            else:
                return None

        metadata = []
        while True:
            key = self.unpack('B')
            if key == 255:
                return metadata
            ty = self.unpack('B')
            if   ty == 0:  val = self.unpack('B')
            elif ty == 1:  val = self.unpack_varint()
            elif ty == 2:  val = self.unpack('f')
            elif ty == 3:  val = self.unpack_string()
            elif ty == 4:  val = self.unpack_chat()
            elif ty == 5:  val = self.unpack_slot()
            elif ty == 6:  val = self.unpack('?')
            elif ty == 7:  val = self.unpack('fff')
            elif ty == 8:  val = self.unpack_position()
            elif ty == 9:  val = unpack_optional(self.unpack_position)
            elif ty == 10: val = self.unpack_varint()
            elif ty == 11: val = unpack_optional(self.unpack_uuid)
            elif ty == 12: val = self.unpack_varint()
            elif ty == 13: val = self.unpack_nbt()
            else: raise ValueError("Unknown entity metadata type: %d" % ty)
            metadata.append((ty, key, val))


    @classmethod
    def pack(cls, fmt, *fields):
        """
        Pack *fields* into a struct. The format accepted is the same as for
        ``struct.pack()``.
        """

        return struct.pack(">"+fmt, *fields)

    @classmethod
    def pack_string(cls, text):
        """
        Pack a Minecraft string (varint-prefixed utf8).
        """

        text = text.encode("utf-8")
        return cls.pack_varint(len(text), max_bits=16) + text

    @classmethod
    def pack_json(cls, obj):
        """
        Serialize an object to JSON and pack it to a Minecraft string.
        """

        return cls.pack_string(json.dumps(obj))

    @classmethod
    def pack_chat(cls, message):
        """
        Pack a Minecraft chat message.
        """
        from quarry.types import chat
        if not isinstance(message, chat.Message):
            message = chat.Message.from_string(message)
        return message.to_bytes()

    @classmethod
    def pack_varint(cls, number, max_bits=32, signed=False):
        """
        Packs a varint.
        """

        number_min = -1 << (max_bits - 1)
        number_max = +1 << (max_bits - 1)
        if not (number_min <= number < number_max):
            raise ValueError("varint does not fit in range: %d <= %d < %d"
                             % (number_min, number, number_max))

        if number < 0:
            if signed:
                number += 1<<32
            else:
                raise ValueError("varint cannot be negative: %d" % number)

        out = b""
        for i in range(5):
            b = number & 0x7F
            number >>= 7
            out += cls.pack("B", b | (0x80 if number > 0 else 0))
            if number == 0:
                break
        return out

    @classmethod
    def pack_uuid(cls, uuid):
        """
        Packs a UUID.
        """

        return uuid.to_bytes()

    @classmethod
    def pack_position(cls, x, y, z):
        """
        Packs a Position.
        """

        def pack_twos_comp(bits, number):
            if number < 0:
                number = number + (1 << bits)
            return number

        return cls.pack('Q', sum((
            pack_twos_comp(26, x) << 38,
            pack_twos_comp(12, y) << 26,
            pack_twos_comp(26, z))))

    @classmethod
    def pack_slot(cls, id=-1, count=1, damage=0, tag=None):
        """
        Packs a slot.
        """

        if id == -1:
            return cls.pack('h', id)

        return cls.pack('hbh', id, count, damage) + cls.pack_nbt(tag)

    @classmethod
    def pack_nbt(cls, tag=None):
        """
        Packs an NBT tag
        """

        if tag is None:
            # slower but more obvious:
            #   from quarry.types import nbt
            #   tag = nbt.TagRoot({})
            return b"\x00"

        return tag.to_bytes()

    @classmethod
    def pack_chunk_section(cls, blocks, block_lights, sky_lights=None):
        """
        Packs a chunk section. The supplied arguments should be instances of
        ``BlockArray`` and ``LightArray`` from ``quarry.types.chunk``.
        """
        out = Buffer.pack('B', blocks.bits)

        if blocks.palette is None:
            out += Buffer.pack_varint(0)
        else:
            out += Buffer.pack_varint(len(blocks.palette))
            out += b"".join(Buffer.pack_varint(x) for x in blocks.palette)

        out += Buffer.pack_varint(len(blocks.data))
        out += Buffer.pack('Q' * len(blocks.data), *blocks.data)
        out += Buffer.pack('B' * 2048, *block_lights.data)
        if sky_lights:
            out += Buffer.pack('B' * 2048, *sky_lights.data)

        return out

    @classmethod
    def pack_entity_metadata_1_7(cls, metadata):
        """
        Packs entity metadata. Compatible with Minecraft 1.7 and 1.8.
        """
        out = b""
        for ty, key, val in metadata:
            out += cls.pack('B', ty << 5 | key)
            if   ty == 0: out += cls.pack('b', val)
            elif ty == 1: out += cls.pack('h', val)
            elif ty == 2: out += cls.pack('i', val)
            elif ty == 3: out += cls.pack('f', val)
            elif ty == 4: out += cls.pack_string(val)
            elif ty == 5: out += cls.pack_slot(**val)
            elif ty == 6: out += cls.pack('iii', *val)
            elif ty == 7: out += cls.pack('fff', *val)
            else: raise ValueError("Unknown entity metadata type: %d" % ty)
        out += cls.pack('B', 127)
        return out

    @classmethod
    def pack_entity_metadata(cls, metadata):
        """
        Packs entity metadata. Compatible with Minecraft 1.9+.
        """
        def pack_optional(packer, *val):
            if val is None:
                return cls.pack('?', False)
            else:
                return cls.pack('?', True) + packer(*val)

        out = b""
        for ty, key, val in metadata:
            out += cls.pack('BB', key, ty)
            if   ty == 0:  out += cls.pack('B', val)
            elif ty == 1:  out += cls.pack_varint(val)
            elif ty == 2:  out += cls.pack('f', val)
            elif ty == 3:  out += cls.pack_string(val)
            elif ty == 4:  out += cls.pack_chat(val)
            elif ty == 5:  out += cls.pack_slot(**val)
            elif ty == 6:  out += cls.pack('?', val)
            elif ty == 7:  out += cls.pack('fff', *val)
            elif ty == 8:  out += cls.pack_position(*val)
            elif ty == 9:  out += pack_optional(cls.pack_position, *val)
            elif ty == 10: out += cls.pack_varint(val)
            elif ty == 11: out += pack_optional(cls.pack_uuid, val)
            elif ty == 12: out += cls.pack_varint(val)
            elif ty == 13: out += cls.pack_nbt(val)
            else: raise ValueError("Unknown entity metadata type: %d" % ty)
        out += cls.pack('B', 255)
        return out



from collections import Sequence
import math

try:
    xrange
except NameError:
    xrange = range

class _Array(Sequence):
    def __len__(self):
        return 4096


class BlockArray(_Array):
    def __init__(self, data, bits, palette=None):
        self.data = data
        self.bits = bits
        self.palette = palette

    def _encode(self, val):
        return (val[0] << 4) | val[1]

    def _decode(self, val):
        return val >> 4, val & 0x0F

    def __getitem__(self, n):
        if isinstance(n, slice):
            return [self[o] for o in xrange(*n.indices(4096))]
        max_val = (1 << self.bits) - 1
        idx0, off0 = divmod(n * self.bits, 64)
        idx1 = ((n + 1) * self.bits - 1) // 64
        if idx0 == idx1:
            val = max_val & (self.data[idx0] >> off0)
        else:
            val = max_val & (self.data[idx0] >> off0 |
                             self.data[idx1] << (64 - off0))

        if self.palette is not None:
            val = self.palette[val]

        return self._decode(int(val))

    def __setitem__(self, n, val):
        if isinstance(n, slice):
            for o in xrange(*n.indices(4096)):
                self[o] = val[o]
            return

        val = self._encode(val)

        if self.palette is not None:
            try:
                val = self.palette.index(val)
            except ValueError:
                self.repack(reserve=1)

                if self.palette is not None:
                    self.palette.append(val)
                    val = len(self.palette) - 1

        max_val = (1 << self.bits) - 1
        idx0, off0 = divmod(n * self.bits, 64)
        idx1, off1 = ((n + 1) * self.bits - 1) // 64, 64 - off0

        self.data[idx0] &= ~(max_val << off0)
        self.data[idx0] |=       val << off0

        if idx0 != idx1:
            self.data[idx1] &= ~(max_val >> off1)
            self.data[idx1] |=       val >> off1

    def repack(self, reserve=None):
        if reserve is None:
            # Recompute the palette by walking all blocks
            palette = [self._encode(v) for v in set(self)]
            palette_len = len(palette)
        else:
            if self.palette is None:
                # Reserving space in an unpaletted array is a no-op.
                return

            palette = self.palette
            palette_len = len(palette) + reserve

        # Compute new bits
        bits = int(math.ceil(math.log(palette_len, 2)))
        if bits <= 8:
            if bits < 4:
                bits = 4
        else:
            bits = 13
            palette = None

        if self.bits == bits:
            # Nothing to do.
            return

        # Save contents
        values = self[:]

        # Update internals
        self.data[:] = [0] * (64 * bits)
        self.bits = bits
        self.palette = palette

        # Load contents
        self[:] = values


class LightArray(_Array):
    def __init__(self, data):
        self.data = data

    def __getitem__(self, n):
        if isinstance(n, slice):
            return [self[o] for o in xrange(*n.indices(4096))]
        assert isinstance(n, int)
        idx, off = divmod(n, 2)
        if off == 0:
            return self.data[idx] & 0x0F
        else:
            return self.data[idx] >> 4

    def __setitem__(self, n, val):
        assert isinstance(n, int)
        idx, off = divmod(n, 2)
        if off == 0:
            self.data[idx] = (self.data[idx] & 0xF0) | val
        else:
            self.data[idx] = (self.data[idx] & 0x0F) | (val << 4)



import json

from twisted.internet import defer
from twisted.web import client, error
from twisted.python import failure

client.HTTPClientFactory.noisy = False

class HTTPException(Exception):
    def __init__(self, error_type, error_message):
        self.error_type = error_type
        self.error_message = error_message

    def __str__(self):
        return "%s: %s" % (self.error_type, self.error_message)

def request(url, timeout, err_type=Exception, expect_content=False, data=None):
    d0 = defer.Deferred()

    def _callback(data):
        d0.callback(json.loads(data.decode('ascii')))

    def _errback(err):
        if isinstance(err.value, error.Error):
            if err.value.status == b"204":
                if expect_content:
                    err = failure.Failure(err_type(
                        "No Content",
                        "No content was returned by the server"))
                else:
                    d0.callback(None)
                    return
            else:
                data = json.loads(err.value.response.decode('ascii'))
                err = failure.Failure(err_type(
                    data['error'],
                    data['errorMessage']))
        d0.errback(err)

    if data:
        d1 = client.getPage(
            url,
            headers = {b'Content-Type': b'application/json'},
            method = b'POST',
            postdata = json.dumps(data).encode('ascii'))
    else:
        d1 = client.getPage(url)
    d1.addCallbacks(_callback, _errback)

    return d0


from distutils.core import setup

setup(
    name='quarry',
    version='0.9.1',
    author='Barney Gale',
    author_email='barney@barneygale.co.uk',
    url='https://github.com/barneygale/quarry',
    license='MIT',
    description='Minecraft protocol library',
    long_description=open('README.rst').read(),
    install_requires=[
        'twisted >= 13.0.0',
        'cryptography >= 0.9',
        'pyOpenSSL >= 0.15.1',
        'service_identity >= 14.0.0',
    ],
    test_requires=[
        'pytest'
    ],
    packages=[
        "quarry",
        "quarry.data",
        "quarry.net",
        "quarry.types"
    ],
    package_data={'quarry': ['data/packets/*.csv']},
)



from __future__ import absolute_import
import uuid

class UUID(uuid.UUID):
    @classmethod
    def from_hex(cls, hex):
        return UUID(hex=hex)

    @classmethod
    def from_bytes(cls, bytes):
        return UUID(bytes=bytes)

    @classmethod
    def from_offline_player(cls, display_name):
        class FakeNamespace():
            bytes = b'OfflinePlayer:'
        base_uuid = uuid.uuid3(FakeNamespace(), display_name)
        return UUID(bytes=base_uuid.bytes)

    def to_hex(self, with_dashes=True):
        if with_dashes:
            return "%s" % self
        else:
            return self.hex

    def to_bytes(self):
        return self.bytes


from twisted.internet import reactor, protocol, defer
from twisted.python import failure

from quarry.net.protocol import Factory, Protocol, ProtocolError, \
    protocol_modes_inv
from quarry.net import auth, crypto


class ClientProtocol(Protocol):
    """This class represents a connection to a server"""

    recv_direction = "downstream"
    send_direction = "upstream"
    protocol_mode_next = None

    def __init__(self, factory, remote_addr):
        Protocol.__init__(self, factory, remote_addr)

    ### Convenience functions -------------------------------------------------

    def switch_protocol_mode(self, mode):
        self.check_protocol_mode_switch(mode)

        if mode in ("status", "login"):
            # Send handshake
            addr = self.transport.connector.getDestination()
            self.send_packet("handshake",
                self.buff_type.pack_varint(self.protocol_version) +
                self.buff_type.pack_string(addr.host) +
                self.buff_type.pack('H', addr.port) +
                self.buff_type.pack_varint(
                    protocol_modes_inv[self.protocol_mode_next]))

        self.protocol_mode = mode

        if mode == "status":
            # Send status request
            self.send_packet("status_request")

        elif mode == "login":
            # Send login start
            self.send_packet("login_start", self.buff_type.pack_string(
                self.factory.profile.display_name))


    ### Callbacks -------------------------------------------------------------

    def connection_made(self):
        """Called when the connection is established"""
        Protocol.connection_made(self)
        self.switch_protocol_mode(self.protocol_mode_next)

    def auth_ok(self, data):
        """
        Called if the Mojang session server responds to our query. Note that
        this method does not indicate that the server accepted our session; in
        this case :meth:`player_joined` is called.
        """

        # Send encryption response
        p_shared_secret = crypto.encrypt_secret(
            self.public_key,
            self.shared_secret)
        p_verify_token = crypto.encrypt_secret(
            self.public_key,
            self.verify_token)

        # 1.7.x
        if self.protocol_version <= 5:
            pack_array = lambda d: self.buff_type.pack('h', len(d)) + d

        # 1.8.x
        else:
            pack_array = lambda d: self.buff_type.pack_varint(
                len(d), max_bits=16) + d

        self.send_packet("login_encryption_response",
            pack_array(p_shared_secret) +
            pack_array(p_verify_token))

        # Enable encryption
        self.cipher.enable(self.shared_secret)
        self.logger.debug("Encryption enabled")

    def player_joined(self):
        """
        Called when we join the game. If the server is in online mode, this
        means the server accepted our session.
        """
        Protocol.player_joined(self)
        self.logger.info("Joined the game.")

    def player_left(self):
        """Called when we leave the game."""
        Protocol.player_left(self)
        self.logger.info("Left the game.")

    def status_response(self, data):
        """
        If we're connecting in "status" mode, this is called when the server
        sends us information about itself.
        """
        self.close()

    ### Packet handlers -------------------------------------------------------

    def packet_status_response(self, buff):
        p_data = buff.unpack_json()
        self.status_response(p_data)

    def packet_login_disconnect(self, buff):
        p_data = buff.unpack_chat()
        self.logger.warn("Kicked: %s" % p_data)
        self.close()

    def packet_login_encryption_request(self, buff):
        p_server_id    = buff.unpack_string()

        # 1.7.x
        if self.protocol_version <= 5:
            unpack_array = lambda b: b.read(b.unpack('h'))
        # 1.8.x
        else:
            unpack_array = lambda b: b.read(b.unpack_varint(max_bits=16))

        p_public_key   = unpack_array(buff)
        p_verify_token = unpack_array(buff)

        if not self.factory.profile.online:
            raise ProtocolError("Can't log into online-mode server while using"
                                " offline profile")

        self.shared_secret = crypto.make_shared_secret()
        self.public_key = crypto.import_public_key(p_public_key)
        self.verify_token  = p_verify_token

        # make digest
        digest = crypto.make_digest(
            p_server_id.encode('ascii'),
            self.shared_secret,
            p_public_key)

        # do auth
        deferred = self.factory.profile.join(digest)
        deferred.addCallbacks(self.auth_ok, self.auth_failed)

    def packet_login_success(self, buff):
        p_uuid = buff.unpack_string()
        p_display_name = buff.unpack_string()

        self.switch_protocol_mode("play")
        self.player_joined()

    def packet_login_set_compression(self, buff):
        self.set_compression(buff.unpack_varint())

    def packet_set_compression(self, buff):
        self.set_compression(buff.unpack_varint())

    packet_disconnect = packet_login_disconnect


class SpawningClientProtocol(ClientProtocol):
    spawned = False

    def __init__(self, factory, remote_addr):
        # x, y, z, yaw, pitch
        self.pos_look = [0, 0, 0, 0, 0]

        super(SpawningClientProtocol, self).__init__(factory, remote_addr)

    # Send a 'player' packet every tick
    def update_player_inc(self):
        self.send_packet("player", self.buff_type.pack('?', True))

    # Sent a 'player position and look' packet every 20 ticks
    def update_player_full(self):
        self.send_packet("player_position_and_look", self.buff_type.pack('dddff?',
            self.pos_look[0],
            self.pos_look[1],
            self.pos_look[2],
            self.pos_look[3],
            self.pos_look[4],
            True))

    def packet_player_position_and_look(self, buff):
        p_pos_look = buff.unpack('dddff')

        # 1.7.x
        if self.protocol_version <= 5:
            p_on_ground = buff.unpack('?')
            self.pos_look = p_pos_look

        # 1.8.x
        else:
            p_flags = buff.unpack('B')

            for i in range(5):
                if p_flags & (1 << i):
                    self.pos_look[i] += p_pos_look[i]
                else:
                    self.pos_look[i] = p_pos_look[i]

            # 1.9.x
            if self.protocol_version > 47:
                teleport_id = buff.unpack_varint()

        # Send Player Position And Look

        # 1.7.x
        if self.protocol_version <= 5:
            self.send_packet("player_position_and_look", self.buff_type.pack(
                'ddddff?',
                self.pos_look[0],
                self.pos_look[1] - 1.62,
                self.pos_look[1],
                self.pos_look[2],
                self.pos_look[3],
                self.pos_look[4],
                True))

        # 1.8.x
        elif self.protocol_version <= 47:
            self.send_packet("player_position_and_look", self.buff_type.pack(
                'dddff?',
                self.pos_look[0],
                self.pos_look[1],
                self.pos_look[2],
                self.pos_look[3],
                self.pos_look[4],
                True))

        # 1.9.x
        else:
            self.send_packet("teleport_confirm", self.buff_type.pack_varint(
                teleport_id))

        if not self.spawned:
            self.ticker.add_loop(1, self.update_player_inc)
            self.ticker.add_loop(20, self.update_player_full)
            self.spawned = True


class ClientFactory(Factory, protocol.ClientFactory):
    protocol = ClientProtocol

    def __init__(self, profile=None):
        if profile is None:
            profile = auth.OfflineProfile()
        self.profile = profile

    def connect(self, host, port=25565, protocol_mode_next="login",
                protocol_version=0):

        if protocol_mode_next == "status" or protocol_version > 0:
            self.protocol.protocol_mode_next = protocol_mode_next
            if protocol_version > 0:
                self.protocol.protocol_version = protocol_version
            reactor.connectTCP(host, port, self, self.connection_timeout)
            return defer.succeed(self)
        else:
            d0 = defer.Deferred()
            factory = ClientFactory()
            class PingProtocol(factory.protocol):
                def status_response(s, data):
                    s.close()
                    detected_version = int(data["version"]["protocol"])
                    if detected_version in self.minecraft_versions:
                        d1 = self.connect(host, port, protocol_mode_next,
                                     detected_version)
                        d1.chainDeferred(d0)
                    else:
                        d0.errback(failure.Failure(ProtocolError(
                            "Unsupported protocol version: %d"
                            % detected_version)))

            factory.protocol = PingProtocol
            factory.connect(host, port, "status")
            return d0
